@@ -1,4 +1,5 @@
 import re
+import os
 import sys
 import time
 import datetime
@@ -7,6 +8,7 @@ from decimal import Decimal
 import threading
 
 import electrum
+from electrum.bitcoin import TYPE_ADDRESS
 from electrum import WalletStorage, Wallet
 from electrum_gui.kivy.i18n import _
 from electrum.contacts import Contacts
@@ -25,7 +27,7 @@ from kivy.properties import (OptionProperty, AliasProperty, ObjectProperty,
 from kivy.cache import Cache
 from kivy.clock import Clock
 from kivy.factory import Factory
-from kivy.metrics import inch, metrics
+from kivy.metrics import inch
 from kivy.lang import Builder
 
 # lazy imports for factory so that widgets can be used in kv
@@ -36,7 +38,6 @@ Factory.register('OutputList', module='electrum_gui.kivy.uix.dialogs')
 Factory.register('OutputItem', module='electrum_gui.kivy.uix.dialogs')
 
 
-
 #from kivy.core.window import Window
 #Window.softinput_mode = 'below_target'
 
@@ -53,8 +54,6 @@ Cache.register('electrum_widgets', timeout=0)
 from kivy.uix.screenmanager import Screen
 from kivy.uix.tabbedpanel import TabbedPanel
 from kivy.uix.label import Label
-from kivy.uix.checkbox import CheckBox
-from kivy.uix.switch import Switch
 from kivy.core.clipboard import Clipboard
 
 Factory.register('TabbedCarousel', module='electrum_gui.kivy.uix.screens')
@@ -99,7 +98,7 @@ class ElectrumWindow(App):
     def on_history(self, d):
         #Logger.info("on_history")
         if self.history_screen:
-            self.history_screen.update()
+            Clock.schedule_once(lambda dt: self.history_screen.update())
 
     def _get_bu(self):
         return self.electrum_config.get('base_unit', 'mBTC')
@@ -180,20 +179,14 @@ class ElectrumWindow(App):
     :data:`ui_mode` is a read only `AliasProperty` Defaults to 'phone'
     '''
 
-    wallet = ObjectProperty(None)
-    '''Holds the electrum wallet
-
-    :attr:`wallet` is a `ObjectProperty` defaults to None.
-    '''
-
     def __init__(self, **kwargs):
         # initialize variables
         self._clipboard = Clipboard
         self.info_bubble = None
-        self.qrscanner = None
         self.nfcscanner = None
         self.tabs = None
         self.is_exit = False
+        self.wallet = None
 
         super(ElectrumWindow, self).__init__(**kwargs)
 
@@ -204,8 +197,8 @@ class ElectrumWindow(App):
         self.plugins = kwargs.get('plugins', [])
 
         self.gui_object = kwargs.get('gui_object', None)
+        self.daemon = self.gui_object.daemon
 
-        #self.config = self.gui_object.config
         self.contacts = Contacts(self.electrum_config)
         self.invoices = InvoiceStore(self.electrum_config)
 
@@ -214,9 +207,12 @@ class ElectrumWindow(App):
             Clock.create_trigger(self.update_wallet, .5)
         self._trigger_update_status =\
             Clock.create_trigger(self.update_status, .5)
-        self._trigger_notify_transactions = \
-            Clock.create_trigger(self.notify_transactions, 5)
+        # cached dialogs
+        self._settings_dialog = None
+        self._password_dialog = None
 
+    def wallet_name(self):
+        return os.path.basename(self.wallet.storage.path) if self.wallet else ' '
 
     def on_pr(self, pr):
         if pr.verify(self.contacts):
@@ -238,15 +234,20 @@ class ElectrumWindow(App):
             self.send_screen.do_clear()
 
     def on_qr(self, data):
+        from electrum.bitcoin import base_decode, is_address
+        data = data.strip()
+        if is_address(data):
+            self.set_URI(data)
+            return
         if data.startswith('bitcoin:'):
             self.set_URI(data)
             return
         # try to decode transaction
-        from electrum.bitcoin import base_decode
         from electrum.transaction import Transaction
         try:
             text = base_decode(data, None, base=43).encode('hex')
             tx = Transaction(text)
+            tx.deserialize()
         except:
             tx = None
         if tx:
@@ -307,7 +308,7 @@ class ElectrumWindow(App):
             return
         from jnius import autoclass
         from android import activity
-        PythonActivity = autoclass('org.renpy.android.PythonActivity')
+        PythonActivity = autoclass('org.kivy.android.PythonActivity')
         Intent = autoclass('android.content.Intent')
         intent = Intent("com.google.zxing.client.android.SCAN")
         intent.putExtra("SCAN_MODE", "QR_CODE_MODE")
@@ -325,6 +326,41 @@ class ElectrumWindow(App):
         except:
             self.show_error(_('Could not start Barcode Scanner.') + ' ' + _('Please install the Barcode Scanner app from ZXing'))
 
+    def scan_qr_zxing(self, on_complete):
+        # uses zxing embedded lib
+        if platform != 'android':
+            return
+        from jnius import autoclass
+        from android import activity
+        PythonActivity = autoclass('org.kivy.android.PythonActivity')
+        IntentIntegrator = autoclass('com.google.zxing.integration.android.IntentIntegrator')
+        integrator = IntentIntegrator(PythonActivity.mActivity)
+        def on_qr_result(requestCode, resultCode, intent):
+            if requestCode == 0:
+                if resultCode == -1: # RESULT_OK:
+                    contents = intent.getStringExtra("SCAN_RESULT")
+                    if intent.getStringExtra("SCAN_RESULT_FORMAT") == 'QR_CODE':
+                        on_complete(contents)
+                    else:
+                        self.show_error("wrong format " + intent.getStringExtra("SCAN_RESULT_FORMAT"))
+        activity.bind(on_activity_result=on_qr_result)
+        integrator.initiateScan()
+
+    def do_share(self, data, title):
+        if platform != 'android':
+            return
+        from jnius import autoclass, cast
+        JS = autoclass('java.lang.String')
+        Intent = autoclass('android.content.Intent')
+        sendIntent = Intent()
+        sendIntent.setAction(Intent.ACTION_SEND)
+        sendIntent.setType("text/plain")
+        sendIntent.putExtra(Intent.EXTRA_TEXT, JS(data))
+        PythonActivity = autoclass('org.kivy.android.PythonActivity')
+        currentActivity = cast('android.app.Activity', PythonActivity.mActivity)
+        it = Intent.createChooser(sendIntent, cast('java.lang.CharSequence', JS(title)))
+        currentActivity.startActivity(it)
+
     def build(self):
         return Builder.load_file('gui/kivy/main.kv')
 
@@ -332,7 +368,7 @@ class ElectrumWindow(App):
         if platform == 'android':
             # move activity to back
             from jnius import autoclass
-            python_act = autoclass('org.renpy.android.PythonActivity')
+            python_act = autoclass('org.kivy.android.PythonActivity')
             mActivity = python_act.mActivity
             mActivity.moveTaskToBack(True)
 
@@ -341,7 +377,6 @@ class ElectrumWindow(App):
         '''
         import time
         Logger.info('Time to on_start: {} <<<<<<<<'.format(time.clock()))
-        Logger.info("dpi: {} {}".format(metrics.dpi, metrics.dpi_rounded))
         win = Window
         win.bind(size=self.on_size, on_keyboard=self.on_keyboard)
         win.bind(on_key_down=self.on_key_down)
@@ -357,7 +392,7 @@ class ElectrumWindow(App):
         if platform == 'android':
             from android import activity
             from jnius import autoclass
-            PythonActivity = autoclass('org.renpy.android.PythonActivity')
+            PythonActivity = autoclass('org.kivy.android.PythonActivity')
             mactivity = PythonActivity.mActivity
             self.on_new_intent(mactivity.getIntent())
             activity.bind(on_new_intent=self.on_new_intent)
@@ -367,33 +402,42 @@ class ElectrumWindow(App):
         if uri:
             self.set_URI(uri)
 
-    def load_wallet_by_name(self, wallet_path):
-        if not wallet_path:
-            return
-        config = self.electrum_config
-        storage = WalletStorage(wallet_path)
-        Logger.info('Electrum: Check for existing wallet')
-        if storage.file_exists:
-            wallet = Wallet(storage)
-            action = wallet.get_action()
-        else:
-            action = 'new'
-        if action is not None:
-            # start installation wizard
-            Logger.debug('Electrum: Wallet not found. Launching install wizard')
-            wizard = Factory.InstallWizard(config, self.network, storage)
-            wizard.bind(on_wizard_complete=lambda instance, wallet: self.load_wallet(wallet))
-            wizard.run(action)
+    def get_wallet_path(self):
+        if self.wallet:
+            return self.wallet.storage.path
         else:
+            return ''
+
+    def on_wizard_complete(self, instance, wallet):
+        if wallet:
+            wallet.start_threads(self.daemon.network)
+            self.daemon.add_wallet(wallet)
             self.load_wallet(wallet)
         self.on_resume()
 
+    def load_wallet_by_name(self, path):
+        if not path:
+            return
+        wallet = self.daemon.load_wallet(path)
+        if wallet:
+            if wallet != self.wallet:
+                self.stop_wallet()
+                self.load_wallet(wallet)
+                self.on_resume()
+        else:
+            Logger.debug('Electrum: Wallet not found. Launching install wizard')
+            wizard = Factory.InstallWizard(self.electrum_config, path)
+            wizard.bind(on_wizard_complete=self.on_wizard_complete)
+            action = wizard.storage.get_action()
+            wizard.run(action)
+
     def on_stop(self):
         self.stop_wallet()
 
     def stop_wallet(self):
         if self.wallet:
-            self.wallet.stop_threads()
+            self.daemon.stop_wallet(self.wallet.storage.path)
+            self.wallet = None
 
     def on_key_down(self, instance, key, keycode, codepoint, modifiers):
         if 'ctrl' in modifiers:
@@ -419,17 +463,21 @@ class ElectrumWindow(App):
             self.is_exit = True
             self.show_info(_('Press again to exit'))
             return True
-        self.is_exit = False
         # override settings button
         if key in (319, 282): #f1/settings button on android
             #self.gui.main_gui.toggle_settings(self)
             return True
 
+    def settings_dialog(self):
+        if self._settings_dialog is None:
+            from uix.dialogs.settings import SettingsDialog
+            self._settings_dialog = SettingsDialog(self)
+        self._settings_dialog.update()
+        self._settings_dialog.open()
+
     def popup_dialog(self, name):
         if name == 'settings':
-            from uix.dialogs.settings import SettingsDialog
-            d = SettingsDialog(self)
-            d.open()
+            self.settings_dialog()
         elif name == 'wallets':
             from uix.dialogs.wallets import WalletDialog
             d = WalletDialog()
@@ -473,10 +521,9 @@ class ElectrumWindow(App):
 
         # connect callbacks
         if self.network:
-            interests = ['updated', 'status', 'new_transaction']
+            interests = ['updated', 'status', 'new_transaction', 'verified']
             self.network.register_callback(self.on_network, interests)
 
-        #self.wallet = None
         self.tabs = self.root.ids['tabs']
 
     def on_network(self, event, *args):
@@ -485,44 +532,49 @@ class ElectrumWindow(App):
         elif event == 'status':
             self._trigger_update_status()
         elif event == 'new_transaction':
-            self._trigger_notify_transactions(*args)
+            self._trigger_update_wallet()
+        elif event == 'verified':
+            self._trigger_update_wallet()
 
     @profiler
     def load_wallet(self, wallet):
-        self.stop_wallet()
         self.wallet = wallet
-        self.wallet.start_threads(self.network)
-        self.current_account = self.wallet.storage.get('current_account', None)
         self.update_wallet()
         # Once GUI has been initialized check if we want to announce something
         # since the callback has been called before the GUI was initialized
+        if self.receive_screen:
+            self.receive_screen.clear()
         self.update_tabs()
-        self.notify_transactions()
         run_hook('load_wallet', wallet, self)
 
     def update_status(self, *dt):
         if not self.wallet:
+            self.status = _("No Wallet")
             return
         if self.network is None or not self.network.is_running():
-            self.status = _("Offline")
+            status = _("Offline")
         elif self.network.is_connected():
             server_height = self.network.get_server_height()
             server_lag = self.network.get_local_height() - server_height
             if not self.wallet.up_to_date or server_height == 0:
-                self.status = _("Synchronizing...")
+                status = _("Synchronizing...")
             elif server_lag > 1:
-                self.status = _("Server lagging (%d blocks)"%server_lag)
+                status = _("Server lagging (%d blocks)"%server_lag)
             else:
-                c, u, x = self.wallet.get_account_balance(self.current_account)
+                c, u, x = self.wallet.get_balance()
                 text = self.format_amount(c+x+u)
-                self.status = str(text.strip() + ' ' + self.base_unit)
+                status = str(text.strip() + ' ' + self.base_unit)
         else:
-            self.status = _("Not connected")
+            status = _("Not connected")
+        n = self.wallet.basename()
+        self.status = '[size=15dp]%s[/size]\n%s' %(n, status) if n !='default_wallet' else status
 
     def get_max_amount(self):
         inputs = self.wallet.get_spendable_coins(None)
         addr = str(self.send_screen.screen.address) or self.wallet.dummy_address()
-        amount, fee = self.wallet.get_max_amount(self.electrum_config, inputs, addr, None)
+        outputs = [(TYPE_ADDRESS, addr, '!')]
+        tx = self.wallet.make_unsigned_transaction(inputs, outputs, self.electrum_config)
+        amount = tx.output_value()
         return format_satoshis_plain(amount, self.decimal_point())
 
     def format_amount(self, x, is_diff=False, whitespaces=False):
@@ -534,50 +586,14 @@ class ElectrumWindow(App):
     @profiler
     def update_wallet(self, *dt):
         self._trigger_update_status()
-        #if self.wallet.up_to_date or not self.network or not self.network.is_connected():
-        self.update_tabs()
-
-    @profiler
-    def notify_transactions(self, *dt):
-        if not self.network or not self.network.is_connected():
-            return
-        # temporarily disabled for merge
-        return
-        iface = self.network
-        ptfn = iface.pending_transactions_for_notifications
-        if len(ptfn) > 0:
-            # Combine the transactions if there are more then three
-            tx_amount = len(ptfn)
-            if(tx_amount >= 3):
-                total_amount = 0
-                for tx in ptfn:
-                    is_relevant, is_mine, v, fee = self.wallet.get_tx_value(tx)
-                    if(v > 0):
-                        total_amount += v
-                self.notify(_("{txs}s new transactions received. Total amount"
-                              "received in the new transactions {amount}s"
-                              "{unit}s").format(txs=tx_amount,
-                                    amount=self.format_amount(total_amount),
-                                    unit=self.base_unit()))
-
-                iface.pending_transactions_for_notifications = []
-            else:
-              for tx in iface.pending_transactions_for_notifications:
-                  if tx:
-                      iface.pending_transactions_for_notifications.remove(tx)
-                      is_relevant, is_mine, v, fee = self.wallet.get_tx_value(tx)
-                      if(v > 0):
-                          self.notify(
-                              _("{txs} new transaction received. {amount} {unit}").
-                              format(txs=tx_amount, amount=self.format_amount(v),
-                                     unit=self.base_unit))
+        if self.wallet and (self.wallet.up_to_date or not self.network or not self.network.is_connected()):
+            self.update_tabs()
 
     def notify(self, message):
         try:
             global notification, os
             if not notification:
                 from plyer import notification
-                import os
             icon = (os.path.dirname(os.path.realpath(__file__))
                     + '/../../' + self.icon)
             notification.notify('Electrum', message,
@@ -587,17 +603,16 @@ class ElectrumWindow(App):
 
     def on_pause(self):
         # pause nfc
-        if self.qrscanner:
-            self.qrscanner.stop()
         if self.nfcscanner:
             self.nfcscanner.nfc_disable()
         return True
 
     def on_resume(self):
-        if self.qrscanner and qrscanner.get_parent_window():
-            self.qrscanner.start()
         if self.nfcscanner:
             self.nfcscanner.nfc_enable()
+        # workaround p4a bug:
+        # show an empty info bubble, to refresh the display
+        self.show_info_bubble('', duration=0.1, pos=(0,0), width=1, arrow_pos=None)
 
     def on_size(self, instance, value):
         width, height = value
@@ -738,47 +753,89 @@ class ElectrumWindow(App):
         popup.open()
 
     def protected(self, msg, f, args):
-        if self.wallet.use_encryption:
+        if self.wallet.has_password():
             self.password_dialog(msg, f, args)
         else:
             apply(f, args + (None,))
 
+    def delete_wallet(self):
+        from uix.dialogs.question import Question
+        basename = os.path.basename(self.wallet.storage.path)
+        d = Question(_('Delete wallet?') + '\n' + basename, self._delete_wallet)
+        d.open()
+
+    def _delete_wallet(self, b):
+        if b:
+            basename = os.path.basename(self.wallet.storage.path)
+            self.protected(_("Enter your PIN code to confirm deletion of %s") % basename, self.__delete_wallet, ())
+
+    def __delete_wallet(self, pw):
+        wallet_path = self.get_wallet_path()
+        dirname = os.path.dirname(wallet_path)
+        basename = os.path.basename(wallet_path)
+        if self.wallet.has_password():
+            try:
+                self.wallet.check_password(pw)
+            except:
+                self.show_error("Invalid PIN")
+                return
+        self.stop_wallet()
+        os.unlink(wallet_path)
+        self.show_error("Wallet removed:" + basename)
+        d = os.listdir(dirname)
+        name = 'default_wallet'
+        new_path = os.path.join(dirname, name)
+        self.load_wallet_by_name(new_path)
+
     def show_seed(self, label):
         self.protected(_("Enter your PIN code in order to decrypt your seed"), self._show_seed, (label,))
 
     def _show_seed(self, label, password):
+        if self.wallet.has_password() and password is None:
+            return
+        keystore = self.wallet.keystore
         try:
-            seed = self.wallet.get_seed(password)
+            seed = keystore.get_seed(password)
+            passphrase = keystore.get_passphrase(password)
         except:
             self.show_error("Invalid PIN")
             return
         label.text = _('Seed') + ':\n' + seed
+        if passphrase:
+            label.text += '\n\n' + _('Passphrase') + ': ' + passphrase
 
-    def change_password(self):
-        self.protected(_("Changing PIN code.") + '\n' + _("Enter your current PIN:"), self._change_password, ())
+    def change_password(self, cb):
+        if self.wallet.has_password():
+            self.protected(_("Changing PIN code.") + '\n' + _("Enter your current PIN:"), self._change_password, (cb,))
+        else:
+            self._change_password(cb, None)
 
-    def _change_password(self, old_password):
-        if self.wallet.use_encryption:
+    def _change_password(self, cb, old_password):
+        if self.wallet.has_password():
+            if old_password is None:
+                return
             try:
                 self.wallet.check_password(old_password)
             except InvalidPassword:
                 self.show_error("Invalid PIN")
                 return
-        self.password_dialog(_('Enter new PIN'), self._change_password2, (old_password,))
+        self.password_dialog(_('Enter new PIN'), self._change_password2, (cb, old_password,))
 
-    def _change_password2(self, old_password, new_password):
-        self.password_dialog(_('Confirm new PIN'), self._change_password3, (old_password, new_password))
+    def _change_password2(self, cb, old_password, new_password):
+        self.password_dialog(_('Confirm new PIN'), self._change_password3, (cb, old_password, new_password))
 
-    def _change_password3(self, old_password, new_password, confirmed_password):
+    def _change_password3(self, cb, old_password, new_password, confirmed_password):
         if new_password == confirmed_password:
             self.wallet.update_password(old_password, new_password)
+            cb()
         else:
             self.show_error("PIN numbers do not match")
 
     def password_dialog(self, msg, f, args):
-        from uix.dialogs.password_dialog import PasswordDialog
         def callback(pw):
             Clock.schedule_once(lambda x: apply(f, args + (pw,)), 0.1)
-        popup = PasswordDialog(msg, callback)
-        popup.open()
-
+        if self._password_dialog is None:
+            from uix.dialogs.password_dialog import PasswordDialog
+            self._password_dialog = PasswordDialog()
+        self._password_dialog.init(msg, callback)
+        self._password_dialog.open()
