@@ -4,21 +4,29 @@
 # Electrum - lightweight Bitcoin client
 # Copyright (C) 2011 thomasv@gitorious
 #
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
+# Permission is hereby granted, free of charge, to any person
+# obtaining a copy of this software and associated documentation files
+# (the "Software"), to deal in the Software without restriction,
+# including without limitation the rights to use, copy, modify, merge,
+# publish, distribute, sublicense, and/or sell copies of the Software,
+# and to permit persons to whom the Software is furnished to do so,
+# subject to the following conditions:
 #
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-# GNU General Public License for more details.
+# The above copyright notice and this permission notice shall be
+# included in all copies or substantial portions of the Software.
 #
-# You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+# SOFTWARE.
 
 import hashlib
 import base64
+import os
 import re
 import hmac
 
@@ -30,11 +38,17 @@ import aes
 
 ################################## transactions
 
-MIN_RELAY_TX_FEE = 5000
+FEE_STEP = 10000
 RECOMMENDED_FEE = 50000
 COINBASE_MATURITY = 100
 COIN = 100000000
 
+# supported types of transction outputs
+TYPE_ADDRESS = 0
+TYPE_PUBKEY  = 1
+TYPE_SCRIPT  = 2
+
+
 # AES encryption
 EncodeAES = lambda secret, s: base64.b64encode(aes.encryptData(secret,s))
 DecodeAES = lambda secret, e: aes.decryptData(secret, base64.b64decode(e))
@@ -146,7 +160,7 @@ hmac_sha_512 = lambda x,y: hmac.new(x, y, hashlib.sha512).digest()
 
 def is_new_seed(x, prefix=version.SEED_PREFIX):
     import mnemonic
-    x = mnemonic.prepare_seed(x)
+    x = mnemonic.normalize_text(x)
     s = hmac_sha_512("Seed version", x.encode('utf8')).encode('hex')
     return s.startswith(prefix)
 
@@ -159,16 +173,25 @@ def is_old_seed(seed):
         uses_electrum_words = True
     except Exception:
         uses_electrum_words = False
-
     try:
         seed.decode('hex')
         is_hex = (len(seed) == 32 or len(seed) == 64)
     except Exception:
         is_hex = False
-
     return is_hex or (uses_electrum_words and (len(words) == 12 or len(words) == 24))
 
 
+def seed_type(x):
+    if is_old_seed(x):
+        return 'old'
+    elif is_new_seed(x):
+        return 'standard'
+    elif is_new_seed(x, version.SEED_PREFIX_2FA):
+        return '2fa'
+    return ''
+
+is_seed = lambda x: bool(seed_type(x))
+
 # pywallet openssl private key implementation
 
 def i2o_ECPublicKey(pubkey, compressed=False):
@@ -195,16 +218,13 @@ def i2o_ECPublicKey(pubkey, compressed=False):
 ############ functions from pywallet #####################
 
 def hash_160(public_key):
-    try:
-        md = hashlib.new('ripemd160')
-        md.update(sha256(public_key))
-        return md.digest()
-    except Exception:
-        # not available in Android SL4a
-        import ripemd
-        md = ripemd.new(sha256(public_key))
-        return md.digest()
-
+    if 'ANDROID_DATA' in os.environ:
+        from Crypto.Hash import RIPEMD
+        md = RIPEMD.new()
+    else:
+        md = hashlib.new('ripemd')
+    md.update(sha256(public_key))
+    return md.digest()
 
 def public_key_to_bc_address(public_key):
     h160 = hash_160(public_key)
@@ -307,6 +327,8 @@ def ASecretToSecret(key, addrtype=0):
     vch = DecodeBase58Check(key)
     if vch and vch[0] == chr((addrtype+128)&255):
         return vch[1:]
+    elif is_minikey(key):
+        return minikey_to_private_key(key)
     else:
         return False
 
@@ -362,6 +384,15 @@ def is_address(addr):
         return False
     return addr == hash_160_to_bc_address(h, addrtype)
 
+def is_p2pkh(addr):
+    if is_address(addr):
+        addrtype, h = bc_address_to_hash_160(addr)
+        return addrtype in [0]
+
+def is_p2sh(addr):
+    if is_address(addr):
+        addrtype, h = bc_address_to_hash_160(addr)
+        return addrtype in [5]
 
 def is_private_key(key):
     try:
@@ -373,6 +404,19 @@ def is_private_key(key):
 
 ########### end pywallet functions #######################
 
+def is_minikey(text):
+    # Minikeys are typically 22 or 30 characters, but this routine
+    # permits any length of 20 or more provided the minikey is valid.
+    # A valid minikey must begin with an 'S', be in base58, and when
+    # suffixed with '?' have its SHA256 hash begin with a zero byte.
+    # They are widely used in Casascius physical bitoins.
+    return (len(text) >= 20 and text[0] == 'S'
+            and all(c in __b58chars for c in text)
+            and ord(sha256(text + '?')[0]) == 0)
+
+def minikey_to_private_key(text):
+    return sha256(text)
+
 from ecdsa.ecdsa import curve_secp256k1, generator_secp256k1
 from ecdsa.curves import SECP256k1
 from ecdsa.ellipticcurve import Point
@@ -384,9 +428,17 @@ def msg_magic(message):
     return "\x18Bitcoin Signed Message:\n" + encoded_varint + message
 
 
-def verify_message(address, signature, message):
+def verify_message(address, sig, message):
     try:
-        EC_KEY.verify_message(address, signature, message)
+        public_key, compressed = pubkey_from_signature(sig, message)
+        # check public key using the address
+        pubkey = point_to_ser(public_key.pubkey.point, compressed)
+        addr = public_key_to_bc_address(pubkey)
+        if address != addr:
+            raise Exception("Bad signature")
+        # check message
+        h = Hash(msg_magic(message))
+        public_key.verify_digest(sig[1:], h, sigdecode = ecdsa.util.sigdecode_string)
         return True
     except Exception as e:
         print_error("Verification error: {0}".format(e))
@@ -467,6 +519,22 @@ class MyVerifyingKey(ecdsa.VerifyingKey):
         return klass.from_public_point( Q, curve )
 
 
+def pubkey_from_signature(sig, message):
+    if len(sig) != 65:
+        raise Exception("Wrong encoding")
+    nV = ord(sig[0])
+    if nV < 27 or nV >= 35:
+        raise Exception("Bad encoding")
+    if nV >= 31:
+        compressed = True
+        nV -= 4
+    else:
+        compressed = False
+    recid = nV - 27
+    h = Hash(msg_magic(message))
+    return MyVerifyingKey.from_signature(sig[1:], recid, h, curve = SECP256k1), compressed
+
+
 class MySigningKey(ecdsa.SigningKey):
     """Enforce low S values in signatures"""
 
@@ -498,41 +566,27 @@ class EC_KEY(object):
         assert public_key.verify_digest(signature, msg_hash, sigdecode = ecdsa.util.sigdecode_string)
         return signature
 
-    def sign_message(self, message, compressed, address):
+    def sign_message(self, message, is_compressed):
         signature = self.sign(Hash(msg_magic(message)))
         for i in range(4):
-            sig = chr(27 + i + (4 if compressed else 0)) + signature
+            sig = chr(27 + i + (4 if is_compressed else 0)) + signature
             try:
-                self.verify_message(address, sig, message)
+                self.verify_message(sig, message)
                 return sig
             except Exception:
                 continue
         else:
             raise Exception("error: cannot sign message")
 
-    @classmethod
-    def verify_message(self, address, sig, message):
-        if len(sig) != 65:
-            raise Exception("Wrong encoding")
-        nV = ord(sig[0])
-        if nV < 27 or nV >= 35:
-            raise Exception("Bad encoding")
-        if nV >= 31:
-            compressed = True
-            nV -= 4
-        else:
-            compressed = False
-        recid = nV - 27
 
-        h = Hash(msg_magic(message))
-        public_key = MyVerifyingKey.from_signature(sig[1:], recid, h, curve = SECP256k1)
+    def verify_message(self, sig, message):
+        public_key, compressed = pubkey_from_signature(sig, message)
         # check public key
-        public_key.verify_digest(sig[1:], h, sigdecode = ecdsa.util.sigdecode_string)
-        pubkey = point_to_ser(public_key.pubkey.point, compressed)
-        # check that we get the original signing address
-        addr = public_key_to_bc_address(pubkey)
-        if address != addr:
+        if point_to_ser(public_key.pubkey.point, compressed) != point_to_ser(self.pubkey.point, compressed):
             raise Exception("Bad signature")
+        # check message
+        h = Hash(msg_magic(message))
+        public_key.verify_digest(sig[1:], h, sigdecode = ecdsa.util.sigdecode_string)
 
 
     # ECIES encryption/decryption methods; AES-128-CBC with PKCS7 is used as the cipher; hmac-sha256 is used as the mac
@@ -778,3 +832,10 @@ def bip32_private_key(sequence, k, chain):
     for i in sequence:
         k, chain = CKD_priv(k, chain, i)
     return SecretToASecret(k, True)
+
+
+def xkeys_from_seed(seed, passphrase, derivation):
+    from mnemonic import Mnemonic
+    xprv, xpub = bip32_root(Mnemonic.mnemonic_to_seed(seed, passphrase))
+    xprv, xpub = bip32_private_derivation(xprv, "m/", derivation)
+    return xprv, xpub
